/* LaBrea
 * $RCSfile: labrea.c $
 * $Release: 1.0 $
 * $Revision: 2.0 $
 * $Date: 2001/09/13 15:10:03 $
 * $Author: tliston $
 *
 * A program to create a tar pit for the connection attempts to
 * a block of ip addresses. LaBrea sits and listens for ARP
 * "who-has" requests.  When an ARP request for a particular IP goes
 * unanswered for longer than it's "rate" setting (default: 3 seconds),
 * LaBrea crafts an ARP reply that routes all traffic destined for the IP
 * to a "bogus" MAC address.  LaBrea then listens for TCP/IP traffic
 * routed to that MAC address and then responds to any SYN packet with a
 * SYN/ACK packet that it creates.
 *
 * LaBrea completely ignores any other input.  This, forces the "client"
 * to wait for its TCP stack to time-out the connection.  Because stacks
 * tend to be a bit tenacious about "established" connections, this
 * should bog down the scanner for quite a long while.
 *
 * Many thanks to Donald Smith for suggestions and improvements.
 * Additional thanks to the following beta testers for all their help:
 * Barton Bruce, Ben Curran, Andrew Daviel, Bill Dodd, Becky Pinkard, and
 * Micropterus Salmoides.  Also, many thanks to Matthew Franz for his help
 * in putting together the Trinux package.  Thanks also to Anders Reed Mohn
 * for his suggestions.
 *
 * THIS PROGRAM REQUIRES BOTH LIBNET and LIBPCAP in order to compile.
 *
 * Copyright (C) 2001 Tom Liston <tliston@premmag.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Log: labrea.c $
 * Revision 2.0  2001/09/13  15:10:03  tliston
 * Added the -p option, to capture and hold connection attempts in the
 * "persist" state.
 * Added the -b option to log bandwidth usage of the -p option to syslog.
 * Cleaned up and tightened up the code in several places.
 *
 * Revision 1.6  2001/08/28  16:51:00  tliston
 * Added the "-a" option to turn off the default behavior of answering
 * SYN/ACKs and PINGs.  Told you I'd name it after you Anders... :-)
 *
 * Revision 1.5  2001/08/27  15:42:17  tliston
 * Returned to using random numbers for IDs and ACKs.  Generating random
 * numbers on the fly was too slow, so I removed them.  Now, I generate a
 * list of random numbers on startup and cycle through them.
 * Added a "-V" option to print out version information.
 * Other small changes.
 *
 * Revision 1.4  2001/08/22  21:10:11  tliston
 * Various small bug fixes.
 *
 * Revision 1.3  2001/08/22  16:31:02  tliston
 * Added the "-z" flag.  By default, you need to pass a "-z" to LaBrea
 * to get it to run.  This isn't documented in the Usage() information,
 * and LaBrea will complain and send you to the README file if you try
 * to run it without a "-z".  This is to FORCE people to read the docs.
 *
 * Revision 1.2  2001/08/22  16:05:06  tliston
 * Added the ability to specify "ranges" within the exclusion files.
 * (Just to make Micro happy...)
 * Fixed a bug in the list deallocation that caused a sigfault :-(
 * Added some comments.
 *
 * Revision 1.1  2001/08/21  20:44:23  tliston
 * Initial revision
 *
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#ifdef WIN32
#include "win32/wincompat.h"
#include "win32/pcap.h"
#include "win32/libnet.h"
#include "win32/syslog.h"
#include <signal.h>
#include <ctype.h>
#else
#include <pcap.h>
#include <libnet.h>
#include <syslog.h>
#endif

/* remove this to eliminate the requirement that the user */
/* supply a "-z" to the program to run it.  Make's 'em read */
/* the instructions... */
#define USEZFLAG

/* #define to LOG_LOCAL0 (see /usr/include/syslog.h ) or whatever */
/* FACILITY to get logging other than log_DAEMON (djs) */
#define LOGFAC LOG_DAEMON
/* define to whatever makes sense to you for your system setup */
#define LOGTYPE LOG_WARNING
/* don't mess with these, unless you know what you're doing */
#define ARP_REQUEST  1
#define ARP_REPLY    2
#define CULLTIME     2400    /* 40 minutes */
#define RANDSIZE     100     /* size of our random number array */
                             /* note: this should be an even number */

struct kotb {
  struct kotb *next;
  struct kotb *prev;
  bpf_u_int32 ip;
  time_t time;
  u_char mac[6];
};

int throttlesize = 10, dlsize = 0, star = 0, sflag = 0, nkotbcount = 0, currentrand=0;
int logopen = 0, datalog = 0, rate = 3, addresses, xflag = 0, hflag = 0, aflag = 0;
int avgbw = 0, currentbandwidth = 0, maxbw = 0, pflag = 0, past[5] = {0,0,0,0,0};
int bflag = 0;
char *dev = NULL;
char mybuffer[1024];
char mybuffer2[1024];
struct libnet_link_int *plnk = NULL;
bpf_u_int32 base, myip, randqueue[RANDSIZE];
time_t *time_array = NULL;
pcap_t *pc = NULL;
u_int *addr_array = NULL;
u_char *sha=NULL, *exclusion = NULL,MACaddr[6];
static char texpr[65536];
struct kotb *nkotbstart = NULL, *nkotbend = NULL;

static char rcsid[] = "$Id: labrea.c 2.0 2001/09/13 15:10:03 tliston Exp tliston $";

void cleanup2(char *p){
  while(*p != '\0'){
    if((*p == '\r') || (*p == '\n'))
      *p = ' ';
    p++;
  }
}

void cleanup(char *p){
  char *p2 = p;

  while(*p != '\0'){
    while((*p2 == ' ') || (*p2 == '\r')){
	  p2++;
	}
	*p = *p2;
	if(*p == '\n')
	  *p = '\0';
	else {
	  p2++;
	  p++;
	}
  }
}

void FixArrays(int wait){
  FILE *in = NULL;
  bpf_u_int32 exip, lowip, i;
  int c, tempxflag;
  char *p;

  /* shut off capture */
  tempxflag = xflag;
  xflag = 1;
  if(wait)
    sleep(1);
  /* clear the exclusion array */
  for(c = 0; c < addresses+2; c++)
    exclusion[c] = 0;
  /* clear the ARP time array */
  if(addr_array != NULL){
    for(c = 0; c < addresses+2; c++)
      addr_array[c] = 0;
  }
  if(time_array != NULL){
    for(c = 0; c < addresses+2; c++)
      time_array[c] = 0;
  }
  /* make us some random numbers */
  for(c = 0; c < RANDSIZE; c++)
    randqueue[c] = libnet_get_prand(PRu32);
  /* handle hardexclusion file */
  if(hflag){
#ifdef WIN32
    if((in = fopen("LaBreaHardExclude.cfg", "r")) == NULL){
      sprintf(mybuffer, "LaBreaHardExclude.cfg not found - no hard exclusions");
#else
    if((in = fopen("/etc/LaBreaHardExclude", "r")) == NULL){
      sprintf(mybuffer, "/etc/LaBreaHardExclude not found - no hard exclusions");
#endif
      syslog(LOGTYPE, mybuffer);
    } else {
      while(!feof(in)){
        fgets(mybuffer, 1024, in);
        if((*mybuffer != '#') && (strlen(mybuffer) > 2)){
          cleanup(mybuffer);
          lowip = 0;
          if((p = strstr(mybuffer, "-")) != NULL){
            *p = '\0';
            lowip = ntohl(libnet_name_resolve(mybuffer, 0)) - base;
            p++;
          } else
            p = mybuffer;
          exip = ntohl(libnet_name_resolve(p,0));
          exip -= base;
          if((exip < 0) || (exip > addresses)){
            sprintf(mybuffer2, "HardExclude IP makes no sense: %s - Ignoring\n", mybuffer);
            syslog(LOGTYPE, mybuffer2);
          } else {
            if(lowip > 0){
              for(i = lowip; i <= exip; i++)
                exclusion[i] = 2;
            } else
              exclusion[exip] = 2;
          }
        }
      }
      fclose(in);
    }
  }
  /* handle exclusion file */
#ifdef WIN32
  if((in = fopen("LaBreaExclude.cfg", "r")) == NULL){
    sprintf(mybuffer, "LaBreaExclude.cfg not found - no exclusions");
#else
  if((in = fopen("/etc/LaBreaExclude", "r")) == NULL){
    sprintf(mybuffer, "/etc/LaBreaExclude not found - no exclusions");
#endif
    syslog(LOGTYPE, mybuffer);
  } else {
    while(!feof(in)){
      fgets(mybuffer, 1024, in);
      if((*mybuffer != '#') && (strlen(mybuffer) > 2)){
        cleanup(mybuffer);
        lowip = 0;
        if((p = strstr(mybuffer, "-")) != NULL){
          *p = '\0';
          lowip = ntohl(libnet_name_resolve(mybuffer, 0)) - base;
          p++;
        } else
          p = mybuffer;
        exip = ntohl(libnet_name_resolve(p,0));
        exip -= base;
        if((exip < 0) || (exip > addresses)){
          sprintf(mybuffer2, "Exclusion IP makes no sense: %s - Ignoring", mybuffer);
          syslog(LOGTYPE, mybuffer2);
        } else {
          if(lowip > 0){
            for(i = lowip; i <= exip; i++)
              exclusion[i] = 1;
          } else
            exclusion[exip] = 1;
        }
      }
    }
    fclose(in);
  }
  /* turn on ARP capture -- if it was on */
  xflag = tempxflag;
}

#ifndef WIN32
void ServiceTimer(int sig){
  int i;

  avgbw  = (10 * past[0]) + (5 * past[1]) + past[2] + past[3] + past[4];
  avgbw /= 1080;
  for(i = 1; i < 5; i++){
    past[i] = past[i-1];
  }
  past[0] = currentbandwidth;
  if(bflag){
    sprintf(mybuffer, "Current average bw: %i (bytes/sec)", avgbw);
    syslog(LOGTYPE, mybuffer);
  }
  currentbandwidth = 0;
  alarm(60);
}
#endif

void CleanExit(int sig){
  struct pcap_stat stat;
  struct kotb *p = nkotbstart, *t;

  /* show 'em some stats... */
  if((logopen) && (pc != NULL)){
    if (pcap_stats(pc, &stat) >= 0){
	  sprintf(mybuffer,"%d packets received by filter", stat.ps_recv);
      syslog(LOG_INFO, mybuffer);
      sprintf(mybuffer, "%d packets dropped by kernel",stat.ps_drop);
      syslog(LOG_INFO, mybuffer);
    }
  }
#ifndef WIN32
  if(sig == SIGHUP){
    syslog(LOG_INFO, "Re-initializing...");
    FixArrays(1);
    return;
  }
#endif
  syslog(LOG_INFO, "Exiting...");
  /* free up our arrays */
  if(addr_array != NULL)
    free(addr_array);
  if(time_array != NULL)
    free(time_array);
  if(exclusion != NULL)
    free(exclusion);
  /* free our list memory */
  while(p != NULL){
    t = p;
    p = p->next;
    free(t);
  }
  /* shut down syslog connection */
  if(logopen)
    closelog();
  /* clean up our pcap stuff */
  if(pc != NULL)
    pcap_close(pc);
  /* clean up our libnet stuff */
  if(plnk != NULL)
    libnet_close_link_interface(plnk);
#ifdef WIN32
  WSACleanup();
#endif
  exit(0);
}

struct kotb* allocNKOTB(void){
  struct kotb *p;

  /* allocate a chunk o' memory */
  p = malloc(sizeof(struct kotb));
  /* check that we really got some */
  if(p == NULL){
    syslog(LOGTYPE, "allocNKOTB: malloc error");
    CleanExit(0);
  }
  /* set the "next" pointer to null */
  p->next = NULL;
  /* if this is the first on the list... */
  if(nkotbstart == NULL){
    /* set the list start ptr to this block */
    nkotbstart = p;
    /* set the end ptr to this block too... */
    nkotbend = p;
    /* no previous block */
    p->prev = NULL;
  } else {
    /* if it isn't the first one... set prev */
    /* to the last one on the list */
    p->prev = nkotbend;
    /* and plug this in as the last one */
    nkotbend = p;
  }
  /* bump our count */
  nkotbcount++;
  return p;
}

void killNKOTB(struct kotb *t){
  nkotbcount--;
  /* if we have a prev pointer */
  if(t->prev != NULL)
    /* snap our prev's next to our next */
    t->prev->next = t->next;
  else
    nkotbstart = t->next;
  /* if we have a next pointer */
  if(t->next != NULL)
    /* snap our next's prev to our prev */
    t->next->prev = t->prev;
  else
    nkotbend = t->prev;
  free(t);
}

void Usage(void){
  char bigstring[] =
      "Usage: LaBrea <options> <BPF filter>\n"
      "Options:\n"
#ifdef WIN32
      "       -i interface : set interface, i.e. -i 1\n"
      "       -D           : Lists available interfaces for -i\n"
	  "       -E           : MAC address of the interface (i.e. 00:2f:92:34:fe:01)\n"
	  "                      THIS NEEDS TO BE SUPPLIED IN THE WIN32 VERSION!\n"
      "       -I           : Interface IP address with netmask (i.e. 10.1.0.1/24)\n"
	  "                      THIS NEEDS TO BE SUPPLIED IN THE WIN32 VERSION!\n"
#else
      "       -i interface : set a non-default interface\n"
#endif
      "       -t datasize  : set connection throttling size in bytes (default 10)\n"
      "       -r rate      : set arp timeout rate in seconds (default 3)\n"
      "       -s           : \"safe\" operation in a switched environment\n"
      "       -l           : log activity to syslog\n"
      "       -v           : verbosely log activity to syslog\n"
      "       -F filename  : specify a BPF filter filename (Note 2)\n"
      "       -h           : \"hard\" capture IPs (Note 4)\n"
      "       -x           : disable IP capture\n"
#ifndef WIN32
      "       -m           : user specified netmask (Note 1)\n"
      "       -n           : user specified network number (Note 1)\n"
#endif
      "       -V           : print version information and exit\n"
      "       -a           : Do not respond to SYN/ACKs and PINGs (Note 6)\n"
      "       -p maxrate   : \"persist\" state capture connect attempts (Note 7)\n"
#ifndef WIN32
      "       -b           : log current bandwidth usage to syslog\n"
      "Note 1:\n"
      " The network number and netmask are normally loaded from the interface.\n"
      " If you're using an interface that has no IP, you'll have to provide BOTH of\n"
      " these numbers.  These MUST be correct or BAD things may happen!\n"
#endif
      "Note 2:\n"
      " Connections specified by the BPF filter will also be tarpitted.  These\n"
      " connections MUST be firewalled to DROP inbound packets or this won't work!\n"
      "Note 3:\n"
#ifdef WIN32
      " LaBreaExclude.cfg contains a list of IPs (1/line) that you DON'T want LaBrea\n"
#else
      " /etc/LaBreaExclude contains a list of IPs (1/line) that you DON'T want LaBrea\n"
#endif
      " to capture. You don't need to specify \"active\" IPs.  LaBrea won't capture an\n"
      " IP with a machine on it. This is only for empty IPs that you DON'T want\n"
      " captured.\n"
      "Note 4:\n"
#ifdef WIN32
      " LaBreaHardExclude.cfg should contain an IP list that you NEVER want LaBrea\n"
#else
      " /etc/LaBreaHardExclude should contain an IP list that you NEVER want LaBrea\n"
#endif
      " to hard capture.\n"
      "Note 5:\n"
      " IPs can be specified as either a single address (i.e.: \"192.168.0.4\") or as a\n"
      " range of addresses (i.e.: \"192.168.0.1 - 192.168.0.50\")\n"
      "Note 6:\n"
      " By default, LaBrea \"virtual machines\" respond to an inbound SYN/ACK with a\n"
      " RST and are \"pingable\".  The -a option eliminates this behavior.\n"
      "Note 7:\n"
      " LaBrea will permanently capture connect attempts within the limit of the\n"
      " maximum data rate specified (in bytes/sec).\n";
  puts(bigstring);
  CleanExit(0);
}

#ifndef WIN32
void GoDaemon(void){
  pid_t fs;

  if(getppid() != 1){
    fs = fork();
    if(fs > 0)
    {
	  exit(0);
	}
	 /* parent */
    if(fs < 0){
      fprintf(stderr, "Error starting fork");
      exit(1);
    }
    setsid();
  }
  /* redirect stdin/stdout/stderr to /dev/null */
  close(0);
  close(1);
  close(2);
  open("/dev/null", O_RDWR);
  dup(0);
  dup(0);
  return;
}
#endif

static void hdlr2(u_char* client_data, const struct pcap_pkthdr* pcpkt, const u_char* pktdata)
{
  u_char proto, flags, type, code, *ptr = ((u_char*)pktdata) + dlsize, *link = (u_char*)pktdata;
  u_char *ip_protobegin;
  u_long ip_src, ip_dst;
  u_short sport, dport, tlength;
  u_long ack, seq, ipid;
  struct kotb *p = nkotbstart, *t;
  int lgmat, i, forgetit = 0;
  time_t c_time;
  char dst_addr[16], src_addr[16];
  struct in_addr iaddr;

  proto = *(u_char*)(ptr + 9);
  ip_src = *(u_long*)(ptr + 12);
  ip_dst = *(u_long*)(ptr + 16);
  lgmat = (*(u_char*)ptr) & 0xF;
  lgmat *= 4;
  ip_protobegin = ptr + lgmat;
  /* Before we do anything, check to see if */
  /* we have any "new kids on the block" and */
  /* if this packet has one of them as its */
  /* destination.  Route it to them */
  if(nkotbcount > 0){
    /* current time */
    c_time = time(NULL);
    while(p != NULL){
      if(ip_dst == p->ip){
        /* get the total length of the IP portion */
        tlength = ntohs(*(u_short*)(ptr + 2));
        /* stuff in the "correct" MAC */
        for(i = 0; i < 6; i++)
          link[i] = p->mac[i];
        /* fire it back out */
        libnet_write_link_layer(plnk, dev, link, tlength + dlsize);
        /* set ourselves up to be culled in 40 minutes */
        p->time = c_time + CULLTIME;
        /* after we're done, checking others for culling, go home... */
        forgetit = 1;
      }
      /* cull out those seeing no activity */
      if(p->time <= c_time){
        t = p->next;
        killNKOTB(p);
        p = t;
      } else
        p = p->next;
    }
  }
  /* if this was a packet that we're just sending on... return */
  if(forgetit)
    return;
  switch(proto){
  case IPPROTO_TCP:
    flags = (*(u_char*)(ip_protobegin + 13));
    iaddr.s_addr = ip_src;
    strcpy(src_addr, inet_ntoa(iaddr));
    iaddr.s_addr = ip_dst;
    strcpy(dst_addr, inet_ntoa(iaddr));
    sport = ntohs(*(u_short*)ip_protobegin);
    dport = ntohs(*(u_short*)(ip_protobegin + 2));
    /* Check that SYN is set and ACK isn't */
    if((flags & TH_SYN) && ((flags & TH_ACK) == 0)){
      seq = ntohl(*(u_long*)(ip_protobegin + 4)) + 1;
      ack = ntohl(randqueue[currentrand]);
      ipid = ntohl(randqueue[currentrand+1]);
      if(currentrand == (RANDSIZE - 2))
        currentrand = 0;
      else
        currentrand += 2;
      /* let's build us a TCP SYN/ACK packet... right over top of our inbound packet */
      /* ETHERNET SPECIFIC */
      for(i = 0; i < 6; i++){
        /* move the source MAC to destination MAC in the ether header */
        link[i] = link[i+6];
        /* stuff our MAC in as the source */
        link[i+6] = sha[i];
      }
      libnet_build_ip(20, 0, ipid, 0, 255, IPPROTO_TCP, ip_dst, ip_src, NULL, 0, ptr);
      libnet_build_tcp(dport, sport, ack, seq, TH_SYN | TH_ACK, throttlesize, 0, NULL, 0, ip_protobegin);
      libnet_do_checksum(ptr, IPPROTO_IP, 20);
      libnet_do_checksum(ptr, IPPROTO_TCP, 20);
      libnet_write_link_layer(plnk, dev, link, 40 + dlsize);
      if(datalog){
        sprintf(mybuffer, "Teergrubing: %s %i -> %s %i", src_addr, sport, dst_addr, dport);
        syslog(LOGTYPE, mybuffer);
      }
    } else {
      if(aflag)
        return;
      if((flags & TH_SYN) && (flags & TH_ACK)){
        /* respond to a SYN/ACK w/a RST to mimic a "live" machine */
        seq = ntohl(*(u_long*)(ip_protobegin + 4)) + 1;
        ack = ntohl(randqueue[currentrand]); /* random ACK # - (not anymore... DJS) */
        ipid = ntohl(randqueue[currentrand+1]);
        if(currentrand == (RANDSIZE - 2))
          currentrand = 0;
        else
          currentrand += 2;
        /* let's build us a TCP SYN/ACK packet... right over top of our inbound packet */
        /* ETHERNET SPECIFIC */
        for(i = 0; i < 6; i++){
          /* move the source MAC to destination MAC in the ether header */
          link[i] = link[i+6];
          /* stuff our MAC in as the source */
          link[i+6] = sha[i];
        }
        libnet_build_ip(20, 0, ipid, 0, 255, IPPROTO_TCP, ip_dst, ip_src, NULL, 0, ptr);
        libnet_build_tcp(dport, sport, ack, seq, TH_RST, throttlesize, 0, NULL, 0, ip_protobegin);
        libnet_do_checksum(ptr, IPPROTO_IP, 20);
        libnet_do_checksum(ptr, IPPROTO_TCP, 20);
        libnet_write_link_layer(plnk, dev, link, 40 + dlsize);
      } else {
        /* check for a data packet */
        if((pflag) && (flags & TH_ACK)){
          tlength = ntohs(*(u_short*)(ptr + 2));
          if((avgbw < maxbw) && (tlength == (40 + throttlesize))){
            /* we got us a data packet! */
            /* let's play with their head... ack it with a window size of zero */
            seq = ntohl(*(u_long*)(ip_protobegin + 4)) + throttlesize;
            ack = ntohl(*(u_long*)(ip_protobegin + 8));
            ipid = ntohl(randqueue[currentrand+1]);
            if(currentrand == (RANDSIZE - 2))
              currentrand = 0;
            else
              currentrand += 2;
            /* let's build us a TCP SYN/ACK packet... right over top of our inbound packet */
            /* ETHERNET SPECIFIC */
            for(i = 0; i < 6; i++){
              /* move the source MAC to destination MAC in the ether header */
              link[i] = link[i+6];
              /* stuff our MAC in as the source */
              link[i+6] = sha[i];
            }
            libnet_build_ip(20, 0, ipid, 0, 255, IPPROTO_TCP, ip_dst, ip_src, NULL, 0, ptr);
            libnet_build_tcp(dport, sport, ack, seq, TH_ACK, 0, 0, NULL, 0, ip_protobegin);
            libnet_do_checksum(ptr, IPPROTO_IP, 20);
            libnet_do_checksum(ptr, IPPROTO_TCP, 20);
            libnet_write_link_layer(plnk, dev, link, 40 + dlsize);
          }
          if(tlength == 41){
            /* we got us a window probe packet! */
            /* tell 'em they STILL gotta keep waiting */
            /* ...impatient bastards... */
            seq = ntohl(*(u_long*)(ip_protobegin + 4));
            ack = ntohl(*(u_long*)(ip_protobegin + 8));
            ipid = ntohl(randqueue[currentrand+1]);
            if(currentrand == (RANDSIZE - 2))
              currentrand = 0;
            else
              currentrand += 2;
            /* let's build us a TCP SYN/ACK packet... right over top of our inbound packet */
            /* ETHERNET SPECIFIC */
            for(i = 0; i < 6; i++){
              /* move the source MAC to destination MAC in the ether header */
              link[i] = link[i+6];
              /* stuff our MAC in as the source */
              link[i+6] = sha[i];
            }
            libnet_build_ip(20, 0, ipid, 0, 255, IPPROTO_TCP, ip_dst, ip_src, NULL, 0, ptr);
            libnet_build_tcp(dport, sport, ack, seq, TH_ACK, 0, 0, NULL, 0, ip_protobegin);
            libnet_do_checksum(ptr, IPPROTO_IP, 20);
            libnet_do_checksum(ptr, IPPROTO_TCP, 20);
            libnet_write_link_layer(plnk, dev, link, 40 + dlsize);
            currentbandwidth += 81;
          }
        }
        if(datalog > 1) {
          /* I do this to avoid "last message repeated x times..." in the syslog*/
          if(star){
            sprintf(mybuffer, "Activity: %s %i -> %s %i *", src_addr, sport, dst_addr, dport);
            star = 0;
          } else {
            sprintf(mybuffer, "Activity: %s %i -> %s %i", src_addr, sport, dst_addr, dport);
            star = 1;
          }
          syslog(LOGTYPE, mybuffer);
        }
      }
    }
    break;
  case IPPROTO_ICMP:
    if(aflag)
      return;
    tlength = ntohs(*(u_short*)(ptr + 2));
    type = *(u_char*)(ip_protobegin);
    code = *(u_char*)(ip_protobegin + 1);
    if((type == 8) && (code == 0)){
      for(i = 0; i < 6; i++){
        /* move the source MAC to destination MAC in the ether header */
        link[i] = link[i+6];
        /* stuff our MAC in as the source */
        link[i+6] = sha[i];
      }
      libnet_build_ip(tlength - lgmat, 0, htonl(150), 0, 255, IPPROTO_ICMP, ip_dst, ip_src, NULL, 0, ptr);
      *(u_char*)ip_protobegin = (u_char)0;
      libnet_do_checksum(ptr, IPPROTO_IP, lgmat);
      libnet_do_checksum(ptr, IPPROTO_ICMP, tlength - lgmat);
      libnet_write_link_layer(plnk, dev, link, tlength + dlsize);
    }
    break;
  default:
    break;
  }
}

static void hdlr(u_char* client_data, const struct pcap_pkthdr* pcpkt, const u_char* pktdata){
  int i;
  u_char tmp, *ptr = ((u_char*)pktdata) + dlsize, *link = (u_char*)pktdata;
  u_short op, frametype;
  u_int spa, tpa, offset, inettpa;
  time_t current, addrtime;
  struct kotb *newkid, *p = nkotbstart, *t;
  u_char bogusMAC[6] = {0,0,15,255,255,255};
  u_char bcastMAC[6] = {255,255,255,255,255,255};

  /* Note: Many of these offsets will need to be changed when we */
  /* support another link layer besides Ethernet.  The offsets will */
  /* need to be calculated based on the size of the MAC coded into */
  /* the ARP packet... Not hard, just different... */
  /* ETHERNET SPECIFIC */
  frametype = ntohs(*(u_short*)(link+12));
  /* are we dealing with an IP frame? */
  /* pass it off to our other handler */
  if(frametype == 0x800){
    hdlr2(client_data, pcpkt, pktdata);
    return;
  }
  /* is it an ARP packet? */
  if(frametype != 0x806)
    return;
  /* if IP capture is disabled, return */
  if(xflag)
    return;
  /* are we dealing with IP addresses? */
  if(ntohs(*(u_short*)(ptr+2)) != 0x800)
    return;
  /* arp operation */
  op = ntohs(*((u_short*)(ptr+6)));
  /* source IP */
  spa = ntohl(*((bpf_u_int32*)(ptr+14)));
  /* target IP */
  inettpa = *((bpf_u_int32*)(ptr+24));
  tpa = ntohl(inettpa);
  switch(op){
  case ARP_REQUEST:
    /* calculate our offset from "base" of available IPs */
    offset =  tpa - base;
    /* goofy IP address */
    if((offset < 0) || (offset > addresses)){
      sprintf(mybuffer, "ARP-request - IP address not in netblock: %i.%i.%i.%i",((tpa & 0xFF000000) >> 24),
             ((tpa & 0xFF0000) >> 16),((tpa & 0xFF00) >> 8),(tpa & 0xFF));
      syslog(LOGTYPE, mybuffer);
      return;
    }
    /* ignore our own arps */
    if(spa == ntohl(myip))
      return;
    /* handle exclusions */
    if(exclusion[offset] == 1)
        return;
    /* Gratuitous ARP */
    if(spa == tpa){
      /* if we see a gratuitous arp, we want to */
      /* attempt to gracefully step out of the way */
      /* even if we've "hard" captured this... */
      time_array[offset] = 0;
      exclusion[offset] = 0;
      addr_array[offset] = 0;
      /* new kid on the block... add it to our list... */
      /* but first, check to see if it's there already */
      if(nkotbcount > 0){
        while(p != NULL){
          if(p->ip == inettpa){
            p->time = time(NULL) + CULLTIME;
            return;
          }
          p = p->next;
        }
      }
      newkid = allocNKOTB();
      /* set up a time to cull this from the list */
      newkid->time = time(NULL) + CULLTIME;
      /* copy the IP of the new kid */
      newkid->ip = *((bpf_u_int32*)(ptr+24));
      /* ... an their MAC */
      for(i = 0; i < 6; i++)
        newkid->mac[i] = link[i+6];
      return;
    }
    /* if we've grabbed it, it's ours... otherwise, check it */
    if(exclusion[offset] != 3){
      /* current time in seconds */
      current = time(NULL);
      /* pull the stored time out of the array */
      addrtime = time_array[offset];
      /* this arp request came from someone new... start over */
      if(addr_array[offset] != spa)
        addrtime = 0;
      /* top-end timeout - 1 minute (perhaps this should be a command line parameter?) */
      if(addrtime + 60 <= current)
        addrtime = 0;
      /* if the time is 0, replace it with the current time */
      if(addrtime == 0) {
        time_array[offset] = current;
        addr_array[offset] = spa;
        if(sflag){
          /* here we'll send out an arp request of our own... */
          for(i = 0; i < 6; i++){
            /* move the broadcast MAC into the destination MAC in the ether header */
            link[i] = bcastMAC[i];
            /* stuff our MAC in as the source */
            /* in both the link layer and the arp */
            link[i+6] = sha[i];
            ptr[i+8] = sha[i];
          }
          /* drop our IP in as the sender */
          *(u_int*)(ptr + 14) = myip;
          /* we made it... now, send it! */
          libnet_write_link_layer(plnk, dev, link, 28 + dlsize);
        }
        return;
      }
      /* if there is a time there, and we haven't timed out, leave */
      if(addrtime + rate > current){
        if(sflag){
          /* here we'll send out an arp request of our own... */
          for(i = 0; i < 6; i++){
            /* move the broadcast MAC into the destination MAC in the ether header */
            link[i] = bcastMAC[i];
            /* stuff our MAC in as the source */
            /* in both the link layer and the arp */
            link[i+6] = sha[i];
            ptr[i+8] = sha[i];
          }
          /* drop our IP in as the sender */
          *(u_int*)(ptr + 14) = myip;
          /* we made it... now, send it! */
          libnet_write_link_layer(plnk, dev, link, 28 + dlsize);
        }
        return;
      }
    }
    /* timeout! */
    /* if we're hard capturing, and this address isn't excluded, hard capture it */
    if((hflag) && (exclusion[offset] < 2))
      exclusion[offset] = 3;
    /* check if this IP is on our "new kids" list */
    /* if it is, remove it... */
    p = nkotbstart;
    if(nkotbcount > 0){
      while(p != NULL){
        if(p->ip == inettpa){
          t = p;
          killNKOTB(t);
          p = NULL;
        } else
          p = p->next;
      }
    }
    /* back to zero to start over... keeps us from (hopefully) permanently stomping */
    /* on an IP.  If something ever answers, we should move out of the way...*/
	time_array[offset] = 0;
    addr_array[offset] = 0;
    if(datalog > 1){
      /* tell 'em what we're doing */
      sprintf(mybuffer, "Capturing: %i.%i.%i.%i",((tpa & 0xFF000000) >> 24),((tpa & 0xFF0000) >> 16),
             ((tpa & 0xFF00) >> 8),(tpa & 0xFF));
      syslog(LOGTYPE, mybuffer);
    }
    /* let's build us an arp packet... right over top of the original */
    /* ETHERNET SPECIFIC */
    for(i = 0; i < 6; i++){
      /* move the source MAC to destination MAC in the ether header */
      link[i] = link[i+6];
      /* stuff our MAC in as the source */
      link[i+6] = sha[i];
      /* stuff our bogus MAC into the Target MAC */
      ptr[i+18] = bogusMAC[i];
    }
    /* change op -> arp reply */
    ptr[7] = 2;
    /* swap sender/target */
    for(i = 8; i < 18; i++){
      tmp = ptr[i];
      ptr[i] = ptr[i+10];
      ptr[i+10] = tmp;
    }
    /* we made it... now, send it! */
    libnet_write_link_layer(plnk, dev, link, 28 + dlsize);
    /* All your IP are belong to us! */
    break;
  case ARP_REPLY:
    /* someone replied... knock down the time */
    offset = spa - base;
    if((offset < 0) || (offset > addresses)){
      sprintf(mybuffer, "ARP-reply - IP address not in netblock: %i.%i.%i.%i",((spa & 0xFF000000) >> 24),
             ((spa & 0xFF0000) >> 16),((spa & 0xFF00) >> 8),(spa & 0xFF));
      syslog(LOGTYPE, mybuffer);
      return;
    }
    time_array[offset] = 0;
    addr_array[offset] = 0;
    break;
  default:
	break;
  }
  return;
}

int main(int argc, char **argv)
{
  int c, dlink, usernet = 0, zflag = 0, len, interf;
  char ebuf[PCAP_ERRBUF_SIZE], *p = NULL, *ffname = NULL,*pi,*po,ch,ipaddr[20]="";
  char *pnet = NULL, *pmask = NULL;
#ifndef WIN32
  sigset_t set;
#endif
  struct bpf_program bpfp;
  bpf_u_int32 net, mask;
  FILE *in = NULL;
  char bigstring[] =
    "Unsupported link layer!\n"
    "Please contact the author, Tom Liston (tliston@premmag.com)\n"
    "with link layer information so that support can be added. Exiting...\n";
#ifdef USEZFLAG
  char bigstring2[] =
    "You MUST read the LaBrea.README file!  Don't try to run this\n"
    "program without understanding how it works and what it can do!\n"
    "In the LaBrea.README file, you'll find the command line switch\n"
    "necessary to allow LaBrea to run.\n"
    "LaBrea is *NOT* running.\n";
#endif

#ifndef WIN32
  sigemptyset(&set);
  sigprocmask(SIG_SETMASK, &set, NULL);
#endif
  /* set signals so that we behave nicely */
  signal(SIGTERM, CleanExit);
  signal(SIGINT, CleanExit);
#ifndef WIN32
  signal(SIGQUIT, CleanExit);
  signal(SIGHUP, CleanExit);
  signal(SIGALRM, ServiceTimer);
#else
  init_winsock();
#endif
  libnet_seed_prand();
  /* parse the command line */
  while((c = getopt(argc, argv, "F:r:i:I:E:t:DlvVxsn:m:?hzap:b")) != EOF){
    switch(c){
    case 'x':
      xflag = 1;
      break;
    case 'b':
      bflag = 1;
      break;
    case 'v':
      datalog += 2;
      break;
    case 'z':
      zflag = 1;
      break;
    case 's':
      sflag = 1;
      break;
    case 'a':
      aflag = 1;
      break;
    case 'l':
      datalog += 1;
      break;
    case 'n':
      pnet = optarg;
      break;
    case 'm':
      pmask = optarg;
      break;
    case 'F':
      ffname = optarg;
      break;
    case 'i':
      dev = optarg;
      break;
#ifdef WIN32
    case 'D':
 		if((po=pcap_lookupdev(ebuf))==NULL)
		{	fprintf(stderr, "pcap_lookupdev: %s\n", ebuf);
			CleanExit(0);
		}
		PrintDeviceList(po);
		CleanExit(0);
	  break;
    case 'E':
	  	for(len=12,pi=optarg,po=MACaddr;len>0;pi++,po++,len-=2)	
		{	while(*pi && (*pi==' ' || *pi==':' || *pi=='-'))
				pi++;
			ch=tolower(*pi++)-'0';					
			if(ch>9)								
				ch-=('0'-9);
			*po=ch<<4;								
			ch=tolower(*pi)-'0';					
			if(ch>9)								
				ch-=('0'-9);
			*po|=ch;								
		}
		sha=MACaddr;
      break;
	case 'I':
		strncpy(ipaddr,optarg,18);
		ipaddr[18]=0;
		break;
#endif
    case 't':
      throttlesize = atoi(optarg);
      break;
    case 'r':
      rate = atoi(optarg);
      break;
    case 'h':
      hflag = 1;
      break;
    case '?':
      Usage();
      break;
    case 'p':
      maxbw = atoi(optarg);
      pflag = 1;
      throttlesize = 5;
      if(maxbw < 1)
        Usage();
      break;
    case 'V':
      puts(rcsid);
#ifdef WIN32
	  puts("Win32 port by Frank Knobbe <frank@knobbe.net>");
#endif
      CleanExit(0);
    default:
      Usage();
      break;
    }
  }
#ifdef USEZFLAG
  if(zflag == 0){
    fprintf(stderr, "%s", bigstring2);
    CleanExit(0);
  }
#endif
  /* sanity checks */
#ifdef WIN32
  if(sha == NULL){
    fprintf(stderr, "You must specify the MAC address with -E (i.e. -E 11:22:33:44:55:66)... Exiting.");
    CleanExit(0);
  }
  if(!*ipaddr){
    fprintf(stderr, "You must specify the IP address with -I (i.e -I 10.10.2.1)... Exiting.");
    CleanExit(0);
  }

  if(strchr(ipaddr,'/')==NULL){
    fprintf(stderr, "You must specify the IP address with network mask (i.e -I 10.1.2.1/24)... Exiting.");
    CleanExit(0);
  }


	if(dev==NULL)
		interf=0;
	else
		interf=atoi(dev);

	if((po=pcap_lookupdev(ebuf))==NULL)
    {	fprintf(stderr, "pcap_lookupdev: %s\n", ebuf);
		CleanExit(0);
	}
		
	if(interf==0)
    {	fprintf(stderr, "No interface specified with -i!");
		PrintDeviceList(po);
		CleanExit(0);
	}
	dev=GetAdapterFromList(po,interf);
	if(dev==NULL)
	{	fprintf(stderr,"Invalid or missing interface!");
		PrintDeviceList(po);
		CleanExit(0);
	}
	if(!*(dev+1))
	{	pi=dev+1;
		po=pi;
		while(*++pi)
			*po++= *pi++;
		*po=0;
	}
	puts(dev);
#else

  if((pnet != NULL) && (pmask == NULL)){
    fprintf(stderr, "If you specify the network number, you MUST also specify the netmask (-m)... Exiting.");
    CleanExit(0);
  }
  if((pnet == NULL) && (pmask != NULL)){
    fprintf(stderr, "If you specify the network number, you MUST also specify the netmask (-m)... Exiting.");
    CleanExit(0);
  }
  if((pnet != NULL) && (pmask != NULL))
    usernet = 1;
#endif

  /* create our bpf filter */
  if(!xflag)
    strcpy(texpr, "arp or (ip and ether host 00:00:0F:FF:FF:FF)");
  /* tack any additional stuff onto our BPF filter */
  if (optind < argc) {
    if(ffname != NULL){
      fprintf(stderr, "Command line filter and filter file specified - Ambiguous! Exiting\n");
      CleanExit(0);
    }
    if(!xflag)
      strcat(texpr, " or (");
    p = texpr;
    while(*p != '\0')
      p++;
    for (c = optind; c < argc; c++) {
      strcpy(p, argv[c]);
      while (*p != '\0')
        p++;
      *p++ = ' ';
      *p = '\0';
    }
    if(!xflag)
      strcat(texpr, ")");
  }
  /* handle filter file */
  if(ffname != NULL){
    if((in = fopen(ffname, "r")) == NULL){
      fprintf(stderr, "Unable to open filter file - Exiting!\n");
      CleanExit(0);
    } else {
      if(!xflag)
        strcat(texpr, " or (");
      while(!feof(in)){
        fgets(mybuffer, 1024, in);
        if(*mybuffer != '#'){
          cleanup2(mybuffer);
          strcat(texpr, mybuffer);
        }
      }
      fclose(in);
      if(!xflag)
        strcat(texpr, ")");
    }
  }
  /* set up our device */
#ifdef WIN32

	libnet_win32_init(interf);

	po=strchr(ipaddr,'/');
	*po++ =0;
	net=inet_addr(ipaddr);
	mask=getnetmask(po);
	
	myip=htonl(net);

	net&=mask;

/*	printf("  IP: %x\nMask: %x\n Net: %x",myip,mask,net);
	exit(0);
*/
#else

  if(dev != NULL){
    if((dev = pcap_lookupdev(ebuf)) == NULL){
      fprintf(stderr, "pcap_lookupdev: %s\n", ebuf);
      exit(1);
    }
  }

  if(usernet){
    net = libnet_name_resolve(pnet,0);
    mask = libnet_name_resolve(pmask,0);
  } else {
    /* get info on our device */
    if(pcap_lookupnet(dev, &net, &mask, ebuf) < 0){
      fprintf(stderr, "pcap_lookupnet(%s): %s\n", dev, ebuf);
      exit(1);
    }
  }
#endif
  /* how many possible addresses in our netblock? */
  addresses = ntohl(mask ^ 0xFFFFFFFF);
  /* set our base address */
  base = ntohl(net);
  /* allocate our exclusion array */
  exclusion = malloc(addresses+2);
  if(exclusion == NULL){
    fprintf(stderr, "malloc: exclusion\n");
	CleanExit(0);
  }
  /* if we're going to use it, allocate our array for holding ARP times */
  if(!xflag){
    /* allocate a chunk o' RAM */
    time_array = malloc(sizeof(time_t) * (addresses+2));
    if(time_array == NULL){
      fprintf(stderr, "malloc: time_array\n");
	  CleanExit(0);
    }
    /* allocate a second chunk o' RAM */
    addr_array = malloc(sizeof(time_t) * (addresses+2));
    if(addr_array == NULL){
      fprintf(stderr, "malloc: addr_array\n");
	  CleanExit(0);
    }
  }
  /* set up logging */
  /* setup the facility for logging in the makefile djs */
  openlog(argv[0], 0, LOGFAC);
  logopen = 1;
  /* tell 'em we're here... */
  strcpy(mybuffer, "Initiated on interface ");
  strncat(mybuffer, dev, 10);
  syslog(LOG_NOTICE, mybuffer);
  FixArrays(0);
  /* check out our link interface (so we can determine link layer data size) */
  if((plnk = libnet_open_link_interface(dev, ebuf)) == NULL){
    libnet_error(LIBNET_ERR_FATAL, "libnet_open_link_interface(%s): %s\n",
		 dev, ebuf);
  } else {
    dlsize = plnk->linkoffset;
  }
  if(dlsize != 14){
    fprintf(stderr, "%s", bigstring);
    CleanExit(0);
  }
  /* open our pcap device */
  /* note: is there a performance increase by chopping the snaplen? */
  /* we probably only need the first dlsize + 80 bytes... */
  if((pc = pcap_open_live(dev, 1518, 1, 0, ebuf)) == NULL){
    fprintf(stderr, "pcap_open_live(%s): %s\n", dev, ebuf);
    CleanExit(0);
  }
  /* set up our datalink */
  dlink = pcap_datalink(pc);
  /* compile our BPF filter */
  if(pcap_compile(pc, &bpfp, texpr, 1, mask) < 0){
    fprintf(stderr, "pcap_compile: %s\n", texpr);
    CleanExit(0);
  }
  /* attach our filter to our datalink */
  if(pcap_setfilter(pc, &bpfp) < 0){
    fprintf(stderr, "pcap_setfilter failed\n");
    CleanExit(0);
  }
  /* get our MAC address to use when we create our arp packets */
#ifndef WIN32  /* Windows version does not have this. We supply MAC addr as argument */
  if((sha = (u_char *)libnet_get_hwaddr(plnk, dev, ebuf)) == 0)
    libnet_error(LIBNET_ERR_FATAL, "libnet_get_hwaddr(%s): %s\n", dev, ebuf);
  /* we need our IP address for the same reason... */

  if((myip = libnet_get_ipaddr(plnk, dev, ebuf)) == 0)
    libnet_error(LIBNET_ERR_FATAL, "libnet_get_ipaddr(%s): %s\n", dev, ebuf);
  myip = htonl(myip);

  GoDaemon();
  /* fire off an alarm if we're doing persist state capture */
  if(pflag)
    alarm(60);
#endif
  /* loop! */
  for(;;){
    if(pcap_loop(pc, 1, &hdlr, NULL) < 0){
      syslog(LOGTYPE, "Error in pcap loop - EXITING");
#ifdef WIN32
	  WSACleanup();
#endif
      exit(1);
    }
  }
  return(0);
}
